# encoding: utf-8
#
# added by CK 2021/04/06

# -*- coding: utf-8 -*-
#
# @Date: 2020-10-26
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import functools
import os

import click
from click_default_group import DefaultGroup

from sdsstools.daemonizer import DaemonGroup

from lvmscp.actor import SCPActor


def cli_coro(f):
    """Decorator function that allows defining coroutines with click."""

    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return functools.update_wrapper(wrapper, f)


@click.group(cls=DefaultGroup, default="actor", default_if_no_args=True)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the user configuration file.",
    show_envvar=True,
)
@click.pass_context
def lvmscp(ctx, config_file: str | None = None):
    """LVM SCP actor."""

    ctx.obj = {"config_file": config_file}


@lvmscp.group(cls=DaemonGroup, prog="scp_actor", workdir=os.getcwd())
@click.pass_context
@cli_coro
async def actor(ctx):
    """Runs the actor."""

    default_config_file = os.path.join(os.path.dirname(__file__), "etc/lvmscp.yml")
    config_file = ctx.obj["config_file"] or default_config_file

    print("Configuration file", config_file)

    lvmscp_obj = SCPActor.from_config(config_file)

    await lvmscp_obj.start()
    await lvmscp_obj.run_forever()  # type: ignore


def main():
    lvmscp(auto_envvar_prefix="LVMSCP")


if __name__ == "__main__":
    main()



# /usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Mingyeong YANG (mingyeong@khu.ac.kr)
# @Date: 2021-03-22
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import os
import pathlib
from contextlib import suppress
from copy import deepcopy

from typing import ClassVar

from archon.actor import ArchonActor
from archon.actor.tools import get_schema
from clu import Command
from sdsstools import read_yaml_file

from lvmscp import config
from lvmscp.controller import SCPController
from lvmscp.delegate import LVMExposeDelegate

from .commands import parser


__all__ = ["SCPActor", "CommandType"]


class SCPActor(ArchonActor):
    """The main actor class of the lvmscp."""

    parser = parser
    BASE_CONFIG: ClassVar[str | dict | None] = config
    DELEGATE_CLASS = LVMExposeDelegate
    CONTROLLER_CLASS = SCPController

    def __init__(self, *args, **kwargs):
        schema = self.merge_schemas(kwargs.pop("schema", None))

        # Just for typing.
        self.controllers: dict[str, SCPController]

        super().__init__(*args, schema=schema, **kwargs)

        assert self.model

        self.emit_status_task: asyncio.Task | None = None

    async def start(self):
        """Starts the actor."""

        start_result = await super().start()

        delay = self.config.get("status_delay", 30.0)
        self.emit_status_task = asyncio.create_task(self.emit_status(delay))

        return start_result

    async def stop(self):
        """Stops the actor and cancels tasks."""

        if self.emit_status_task and not self.emit_status_task.done():
            self.emit_status_task.cancel()
            with suppress(asyncio.CancelledError):
                await self.emit_status_task

        return await super().stop()

    async def emit_status(self, delay: float = 30.0):
        """Emits the status of the controller on a timer."""

        await asyncio.sleep(5)

        while True:
            await Command(
                "status",
                actor=self,
                commander_id=f".{self.name}",
                internal=True,
                write_to_log=False,
            ).parse()

            await asyncio.sleep(delay)

    def merge_schemas(self, scp_schema_path: str | None = None):
        """Merge default schema with SCP one."""

        schema = get_schema()  # Default archon schema.

        if scp_schema_path:
            root_path = pathlib.Path(__file__).absolute().parents[1]
            if not os.path.isabs(scp_schema_path):
                scp_schema_path = os.path.join(str(root_path), scp_schema_path)

            scp_schema = json.loads(open(scp_schema_path, "r").read())

            schema["definitions"].update(scp_schema.get("definitions", {}))
            schema["properties"].update(scp_schema.get("properties", {}))
            schema["patternProperties"].update(scp_schema.get("patternProperties", {}))

            if "additionalProperties" in scp_schema:
                schema["additionalProperties"] = scp_schema["additionalProperties"]

        return schema

    @classmethod
    def from_config(cls, iconfig, *args, **kwargs):
        """Creates an actor from a configuration file."""

        if isinstance(cls.BASE_CONFIG, str):
            cls.BASE_CONFIG = read_yaml_file(cls.BASE_CONFIG)

        if iconfig is None:
            if cls.BASE_CONFIG is None:
                raise RuntimeError("The class does not have a base configuration.")
            config = deepcopy(cls.BASE_CONFIG)
        else:
            if isinstance(cls.BASE_CONFIG, dict):
                config = deepcopy(cls.BASE_CONFIG)
            else:
                config = {}

            if isinstance(iconfig, str):
                iconfig = read_yaml_file(iconfig)

            config.update(iconfig)

        return super(SCPActor, cls).from_config(config, *args, **kwargs)


CommandType = Command[SCPActor]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-05-29
# @Filename: delegate.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING, Any, List, Literal, Tuple

import numpy
from astropy.coordinates import EarthLocation
from astropy.io import fits
from astropy.time import Time
from astropy.utils.iers import conf

from archon.actor import ExposureDelegate
from archon.controller.controller import ArchonController
from sdsstools.time import get_sjd


if TYPE_CHECKING:
    from clu import Command

    from .actor import SCPActor


conf.auto_download = False
conf.iers_degraded_accuracy = "ignore"


class LVMExposeDelegate(ExposureDelegate["SCPActor"]):
    """Expose delegate for LVM."""

    def __init__(self, actor: SCPActor):
        super().__init__(actor)

        self.use_shutter: bool = True
        self.shutter_failed: bool = False

        # Header values to be collected during integration.
        self.header_data: dict[str, Any] = {}

        # Pressure and depth probe data. These data is per CCD/cryostat.
        self.pressure_data: dict[str, float] = {}
        self.depth_data: dict[str, float | str] = {}

        # LCO
        self.location = EarthLocation.from_geodetic(
            lon=-70.70166667,
            lat=-29.00333333,
            height=2282.0,
        )

    def reset(self):
        self.header_data = {}
        self.pressure_data = {}
        self.depth_data = {}

        self.use_shutter = True

        return super().reset()

    async def check_expose(self) -> bool:
        """Performs a series of checks to confirm we can expose."""

        base_checks = await super().check_expose()
        if not base_checks:
            return False

        if self.use_shutter:
            assert self.expose_data
            controllers = self.expose_data.controllers
            jobs_status = []
            for controller in controllers:
                jobs_status.append(self.get_shutter_status(controller.name))

            results = await asyncio.gather(*jobs_status)
            for result in results:
                if result is False:
                    return self.fail("Failed getting shutter status.")
                if result["invalid"] or result["open"]:
                    return self.fail("Some shutters are in an invalid stated or open.")

        return True

    async def shutter(self, open, is_retry=False):
        """Operate the shutter."""

        if not self.use_shutter:
            return True

        assert self.expose_data
        expose_data = self.expose_data

        if expose_data.exposure_time == 0 or expose_data.flavour in ["bias", "dark"]:
            return True

        action = "open" if open else "close"

        self.command.debug(text=f"Moving shutters to {action}.")

        jobs = []
        for controller in self.expose_data.controllers:
            jobs.append(self.move_shutter(controller.name, action))
        results = await asyncio.gather(*jobs, return_exceptions=True)

        if not all(results):
            self.shutter_failed = True
            if is_retry is False:
                self.command.warning(text="Some shutters failed to close. Retrying.")
                await asyncio.sleep(3)
                return await self.shutter(open, is_retry=True)
            else:
                self.command.error("Some shutters failed to move.")
        else:
            self.shutter_failed = False

        if self.shutter_failed:
            if open is True:
                return False
            else:
                self.command.warning(
                    "Shutter failed to close. Reading out exposure and failing."
                )

        return True

    async def readout(
        self,
        command: Command[SCPActor],
        extra_header: dict[str, Any] = {},
        delay_readout: int = 0,
        write: bool = True,
    ):
        """Reads detectors."""

        if self.shutter_failed:
            self.command.warning(
                "Frame was read out but shutter failed to close. "
                "There may be contamination in the image."
            )

        read_result = await super().readout(command, extra_header, delay_readout, write)

        return False if (self.shutter_failed or not read_result) else True

    async def expose_cotasks(self):
        """Grab sensor data when the exposure begins to save time.

        We read the sensors during the exposure to avoid out of date information
        if, for example, the lamps are turned off during readout. However, in
        very short exposures or biases, the cotasks may take longer than the
        exposure itself and readout will have already begun.

        """

        self.command.debug("Grabbing sensor data and system status.")

        assert self.expose_data

        # We expect only on controller in lvmscp.
        controller = self.expose_data.controllers[0]

        tasks = [
            self.get_hartmann_status(controller.name),
            self.get_sensors(controller.name),
            self.get_lamps(),
            self.get_pressure(controller.name),
            self.read_depth_probes(),
            self.get_telescope_info(),
        ]

        await asyncio.gather(*tasks)

        return

    async def post_process(
        self,
        controller: ArchonController,
        hdus: List[fits.PrimaryHDU],
    ) -> Tuple[ArchonController, List[fits.PrimaryHDU]]:
        """Post-process images."""

        self.command.debug(text="Running exposure post-process.")

        now = Time.now()
        now.location = self.location

        for hdu in hdus:
            ccd = str(hdu.header["CCD"])

            hdu.header["LMST"] = round(now.sidereal_time("mean").value, 6)

            # Update header with values collected during integration.
            for key in self.header_data:
                hdu.header[key] = self.header_data[key]

            # Add SDSS MJD.
            hdu.header["SMJD"] = get_sjd("LCO")
            hdu.header["PRESSURE"] = self.pressure_data.get(f"{ccd}_pressure", -999.0)

            depth_camera = self.depth_data.get("camera", "")
            for ch in ["A", "B", "C"]:
                depth = (self.depth_data[ch] if ccd == depth_camera else -999.0,)
                hdu.header[f"DEPTH{ch}"] = depth

        return (controller, hdus)

    async def get_shutter_status(self, spec: str) -> dict | Literal[False]:
        """Returns the status of the shutter for a spectrograph."""

        lvmieb = self.actor.controllers[spec].lvmieb
        cmd = await self.command.send_command(lvmieb, f"shutter status {spec}")
        await cmd

        if cmd.status.did_fail:
            return False

        try:
            return cmd.replies.get(f"{spec}_shutter")
        except KeyError:
            return False

    async def move_shutter(self, spec: str, action: str) -> bool:
        """Opens/closes a shutter."""

        lvmieb = self.actor.controllers[spec].lvmieb
        cmd = await self.command.send_command(lvmieb, f"shutter {action} {spec}")
        await cmd

        return cmd.status.did_succeed

    async def get_hartmann_status(self, spec: str):
        """Returns the status of the hartmann doors."""

        lvmieb = self.actor.controllers[spec].lvmieb

        cmd = await self.command.send_command(
            lvmieb,
            f"hartmann status {spec}",
            time_limit=5,
        )

        try:
            left = 0 if cmd.replies.get(f"{spec}_hartmann_left")["open"] else 1
            right = 0 if cmd.replies.get(f"{spec}_hartmann_right")["open"] else 1
            self.header_data["HARTMANN"] = f"{int(left)} {int(right)}"
        except KeyError:
            self.command.warning(f"{spec}: failed retrieving hartmann door status.")

    async def get_sensors(self, spec: str):
        """Returns the spectrograph temperatures and RHs."""

        lvmieb = self.actor.controllers[spec].lvmieb
        cmd = await self.command.send_command(
            lvmieb,
            f"wago status {spec}",
            time_limit=5,
        )

        try:
            sensors = cmd.replies.get(f"{spec}_sensors")
            self.header_data["LABTEMP"] = sensors.get("t3", -999.0)
            self.header_data["LABHUMID"] = sensors.get("rh3", -999.0)
        except KeyError:
            self.command.warning(f"{spec}: failed retrieving sensor values.")

    async def get_pressure(self, spec: str):
        """Returns the cryostat pressures."""

        lvmieb = self.actor.controllers[spec].lvmieb
        cmd = await self.command.send_command(
            lvmieb,
            f"transducer status {spec}",
            time_limit=5,
        )

        try:
            self.pressure_data = cmd.replies.get("transducer")
        except KeyError:
            self.command.warning(f"{spec}: failed retrieving pressure status.")

    async def read_depth_probes(self):
        """Returns the depth probe measurements."""

        spec_config = list(self.actor.config["controllers"].values())[0]
        lvmieb_name = spec_config.get("lvmieb", "lvmieb")

        cmd = await self.command.send_command(
            lvmieb_name,
            "depth status",
            time_limit=5,
        )

        try:
            self.depth_data = cmd.replies.get("depth")
        except KeyError:
            pass

    async def get_lamps(self):
        """Retrieves lamp information."""

        lvmnps = self.actor.config.get("lvmnps", "lvmnps")
        cmd = await self.command.send_command(
            lvmnps,
            "status",
            time_limit=10,
        )

        # The config file includes the names of the lamps that should be present.
        lamps = self.actor.config.get("lamps", [])

        try:
            status = cmd.replies.get("status")
            for switch in status:
                for outlet in status[switch]:
                    if outlet in lamps:
                        state = "ON" if status[switch][outlet]["state"] == 1 else "OFF"
                        self.header_data[outlet.upper()] = state
        except Exception as err:
            self.command.warning(f"Failed retrieving lamp status: {err}")

    async def get_telescope_info(self):
        """Retrieve telescope information."""

        telescopes = ["sci", "skye", "skyw", "spec"]

        for telescope in telescopes:
            pwi_cmd = await self.command.send_command(
                f"lvm.{telescope}.pwi",
                "status",
                internal=True,
                time_limit=5,
            )
            if pwi_cmd.status.did_fail:
                self.command.warning(f"Failed getting {telescope} PWI status.")
            else:
                pwi_status = pwi_cmd.replies[-1].body

                ra_h: float = pwi_status.get("ra_j2000_hours", -999.0)
                if ra_h > 0:
                    ra_d = ra_h * 15.0
                else:
                    ra_d = ra_h
                self.header_data[f"TE{telescope.upper()}RA"] = numpy.round(ra_d, 6)

                dec = pwi_status.get("dec_j2000_degs", -999.0)
                self.header_data[f"TE{telescope.upper()}DE"] = numpy.round(dec, 6)

                alt = pwi_status.get("altitude_degs", None)
                if alt is not None:
                    airm = numpy.round(1 / numpy.cos(numpy.radians(90 - alt)), 3)
                    self.header_data[f"TE{telescope.upper()}AM"] = airm

            if telescope != "spec":
                km_cmd = await self.command.send_command(
                    f"lvm.{telescope}.km",
                    "status",
                    internal=True,
                    time_limit=5,
                )
                if km_cmd.status.did_fail:
                    self.command.warning(f"Failed getting {telescope} k-mirror status.")
                else:
                    km_position = numpy.round(km_cmd.replies.get("Position"), 2)
                    self.header_data[f"TE{telescope.upper()}KM"] = km_position

            foc_cmd = await self.command.send_command(
                f"lvm.{telescope}.foc",
                "status",
                internal=True,
                time_limit=5,
            )
            if foc_cmd.status.did_fail:
                self.command.warning(f"Failed getting {telescope} focus status.")
            else:
                foc_position = numpy.round(foc_cmd.replies.get("Position"), 2)
                self.header_data[f"TE{telescope.upper()}FO"] = foc_position
